# Dircolors, a Python library for colorizing and formatting filenames like GNU Coreutils'
# ls and dircolors programs.
# Requires python 3.3 or later
#
# Copyright 2019 Allen Wild <allenwild93@gmail.com>
# Modifications copyright 2020 John Skilleter <john@skilleter.org.uk>
# SPDX-License-Identifier: Apache-2.0

""" dircolors, a Python library to colorize filenames based on their type
for terminal use, like GNU ls and dircolors. """

from collections import OrderedDict
from io import StringIO, TextIOBase
import os
import stat

import dc_defaults
import dc_util

__all__ = ['Dircolors']

_CODE_MAP = OrderedDict()

def _init_code_map():
    """ mapping between the key name in the .dircolors file and the two letter
    code found in the LS_COLORS environment variable.
    Used for parsing .dircolors files. """
    # This code is wrapped in a function so we can disable pylint's whitespace check
    # on a limited scope.
    # pylint: disable=bad-whitespace
    _CODE_MAP['RESET']                  = 'rs'
    _CODE_MAP['DIR']                    = 'di'
    _CODE_MAP['LINK']                   = 'ln'
    _CODE_MAP['MULTIHARDLINK']          = 'mh'
    _CODE_MAP['FIFO']                   = 'pi'
    _CODE_MAP['SOCK']                   = 'so'
    _CODE_MAP['DOOR']                   = 'do'
    _CODE_MAP['BLK']                    = 'bd'
    _CODE_MAP['CHR']                    = 'cd'
    _CODE_MAP['ORPHAN']                 = 'or'
    _CODE_MAP['MISSING']                = 'mi'
    _CODE_MAP['SETUID']                 = 'su'
    _CODE_MAP['SETGID']                 = 'sg'
    _CODE_MAP['CAPABILITY']             = 'ca'
    _CODE_MAP['STICKY_OTHER_WRITABLE']  = 'tw'
    _CODE_MAP['OTHER_WRITABLE']         = 'ow'
    _CODE_MAP['STICKY']                 = 'st'
    _CODE_MAP['EXEC']                   = 'ex'

_init_code_map()
del _init_code_map

class Dircolors:
    """ Main dircolors class. Contains a database of formats corresponding to file types,
    modes, and extensions. Use the format() method to check a file and color it appropriately.
    """

    def __init__(self, load=True):
        """ Initialize a Dircolors object. If load=True (the default), then try
        to load dircolors info from the LS_COLORS environment variable.
        If no data is obtained from LS_COLORS, load the defaults.
        If load=False, don't even load defaults. """
        self._loaded = False
        self._codes = OrderedDict()
        self._extensions = OrderedDict()
        if load:
            if not self.load_from_environ():
                self.load_defaults()

    def __bool__(self):
        """ convenience method for checking whether this Dircolors object has loaded a database.
        Can be used like
            d = Dircolors()
            if d:
                d.format(somefile)
        """
        return self._loaded

    @property
    def loaded(self):
        """ return a boolean indicating whether some valid dircolors data has been loaded """
        return self._loaded

    def clear(self):
        """ Clear the loaded data """
        self._loaded = False
        self._codes.clear()
        self._extensions.clear()

    def load_from_lscolors(self, lscolors):
        """ Load the dircolors database from a string in the same format as the LS_COLORS
        environment variable.
        Returns True if data was successfully loaded, False otherwise (e.g. if
        envvar is unset). Regardless, the current database will be cleared """
        self.clear()
        if not lscolors:
            return False

        for item in lscolors.split(':'):
            try:
                code, color = item.split('=', 1)
            except ValueError:
                continue  # no key=value, just ignore
            if code.startswith('*.'):
                self._extensions[code[1:]] = color
            else:
                self._codes[code] = color

        if self._codes or self._extensions:
            self._loaded = True
        return self._loaded

    def load_from_environ(self, envvar='LS_COLORS'):
        """ Load the dircolors database from an environment variable. By default,
        use LS_COLORS like the GNU Coreutils `ls` program.
        Returns True if data was successfully loaded, False otherwise (e.g. if
        envvar is unset). Regardless, the current database will be cleared. """
        return self.load_from_lscolors(os.environ.get(envvar))

    def load_from_dircolors(self, database, strict=False):
        """ Load the dircolors database from a GNU-compatible .dircolors file.
        May raise any of the usual OSError exceptions if filename doesn't exist
        or otherwise can't be read.

        database can be a string representing a filename, or a file-like object
        opened in text mode (i.e. a subclass of io.TextIOBase). To load from the
        contents of a .dircolors file, wrap it in an io.StringIO object.

        If strict is True, raise ValueError on the first unparsed line

        Returns a boolean indicating whether any data was loaded.
        The current database will always be cleared. """
        self.clear()
        if isinstance(database, str):
            file = open(database, 'r')
        elif isinstance(database, TextIOBase):
            file = database
        else:
            raise ValueError('database must be str or io.TextIOBase, not %s' % type(database))

        try:
            for line in file:
                # remove comments and skip empty lines
                line = line.split('#')[0].strip()
                if not line:
                    continue

                # make sure there's two space-separated fields
                split = line.split()
                if len(split) != 2:
                    if strict:
                        raise ValueError('Warning: unable to parse dircolors line "%s"' % line)
                    continue

                key, val = split
                if key == 'TERM':
                    continue  # ignore TERM directives
                elif key in _CODE_MAP:
                    self._codes[_CODE_MAP[key]] = val
                elif key.startswith('.'):
                    self._extensions[key] = val
                elif strict:
                    raise ValueError('Warning: unable to parse dircolors line "%s"' % line)
                # elif not strict, skip

            if self._codes or self._extensions:
                self._loaded = True
            return self._loaded
        finally:
            file.close()

    def load_defaults(self):
        """ Load the default database. """
        self.clear()
        return self.load_from_dircolors(StringIO(dc_defaults.DEFAULT_DIRCOLORS), True)

    def generate_lscolors(self):
        """ Output the database in the format used by the LS_COLORS environment variable. """
        if not self._loaded:
            return ''

        def gen_pairs():
            for pair in self._codes.items():
                yield pair
            for pair in self._extensions.items():
                # change .xyz to *.xyz
                yield '*' + pair[0], pair[1]

        return ':'.join('%s=%s' % pair for pair in gen_pairs())

    def _format_code(self, text, code):
        """ format text with an lscolors code. Return text unmodified if code
        isn't found in the database """
        val = self._codes.get(code, None)
        if val:
            return '\033[%sm%s\033[%sm' % (val, text, self._codes.get('rs', '0'))
        return text

    def _format_ext(self, text, ext):
        """ Format text according to the given file extension.
        ext must have a leading '.'
        text need not actually end in '.ext' """
        val = self._extensions.get(ext, '0')
        if val:
            return '\033[%sm%s\033[%sm' % (val, text, self._codes.get('rs', '0'))
        return text

    def format_mode(self, text, mode):
        """ Format and color the given text based on the given file mode.

        `mode` can be an integer, usually the st_mode field of an os.stat_result
        object obtained from os.stat() or similar function. It can also be an os.stat_result
        object, and the st_mode field will be extracted automatically.

        If mode is zero then the formatting is performed only according to the filename.

        `text` is an arbitrary string which will be colored according to the bits
        set in `mode` and the colors database loaded in this Dircolors object.

        If `mode` represents a symlink, it will be formatted as such with no dereferencing
        (since this function doesn't know the file name) """
        if not self._loaded:
            return text

        if isinstance(mode, int):
            pass
        elif isinstance(mode, os.stat_result):
            mode = mode.st_mode
        else:
            raise ValueError('mode must be int or os.stat_result, not %s' % type(mode))

        if mode:
            if stat.S_ISDIR(mode):
                if (mode & (stat.S_ISVTX | stat.S_IWOTH)) == (stat.S_ISVTX | stat.S_IWOTH):
                    # sticky and world-writable
                    return self._format_code(text, 'tw')
                if mode & stat.S_ISVTX:
                    # sticky but not world-writable
                    return self._format_code(text, 'st')
                if mode & stat.S_IWOTH:
                    # world-writable but not sticky
                    return self._format_code(text, 'ow')
                # normal directory
                return self._format_code(text, 'di')

            # special file?
            # pylint: disable=bad-whitespace
            special_types = (
                (stat.S_IFLNK,  'ln'),  # symlink
                (stat.S_IFIFO,  'pi'),  # pipe (FIFO)
                (stat.S_IFSOCK, 'so'),  # socket
                (stat.S_IFBLK,  'bd'),  # block device
                (stat.S_IFCHR,  'cd'),  # character device
                (stat.S_ISUID,  'su'),  # setuid
                (stat.S_ISGID,  'sg'),  # setgid
            )

            for mask, code in special_types:
                if (mode & mask) == mask:
                    return self._format_code(text, code)

            # executable file?
            if mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH):
                return self._format_code(text, 'ex')

        # regular file, format according to its extension
        _, ext = os.path.splitext(text)
        if ext:
            return self._format_ext(text, ext)
        return text

    def format(self, file, cwd=None, follow_symlinks=False, show_target=False):
        """ Format and color the file given by the name `file`.

        If `cwd` is not None, it should be a string for the directory relative
        to which `file` is looked up, or an integer representing a directory
        descriptor (usually from `os.open()`).

        If the file does not exist, it is formatted according to the filename only.

        Use follow_symlinks to dereference symlinks entirely.
        Use show_target=True with follow_symlinks=False to format both the link name
        and its target in the format:
            linkname -> target
        With linkname formatted as a link color, and the link target formatted as its respective
        type. If the link target is another link, it will not be recursively dereferenced. """

        if not self.loaded:
            return file

        if os.path.exists(file):
            try:
                statbuf = dc_util.stat_at(file, cwd, follow_symlinks)
            except OSError as e:
                return '%s [Error stat-ing: %s]' % (file, e.strerror)

            mode = statbuf.st_mode
            if (not follow_symlinks) and show_target and stat.S_ISLNK(mode):
                target_path = dc_util.readlink_at(file, cwd)
                try:
                    dc_util.stat_at(target_path, cwd)  # check for broken link
                    target = self.format(target_path, cwd, False, False)
                except OSError:
                    # format as "orphan"
                    target = self._format_code(target_path, 'or') + ' [broken link]'
                return self._format_code(file, 'ln') + ' -> ' + target
        else:
            mode = 0

        return self.format_mode(file, mode)



#! /usr/bin/env python3

################################################################################
""" Functions for making log files and similar more easily viewed and/or compared

    Defaults to making things readable on a light background but has the
    option to make things readable on a dark background or for removing
    colours altogether.

    Has functions for removing times, AWS ID values, colours, SHA1 values,
    Also function for making ANSI codes more readable on light or dark backgrounds

    Currently this is very cude and needs to be a lot more cleverer.

    TODO: Handle multiple colour changes in a single ANSI sequence
    TODO: Take account of the background colour when determining how to change the foreground colour
    TODO: More colour conversions (currently only converts bright yellow on white to dark yellow)
    TODO: Handle 256 colour codes
    TODO: More time formats
"""
################################################################################

import re

################################################################################
# Regular expressions

# Match an ANSI colour control sequence

ANSI_REGEXES = [
    re.compile(r'\x1b\[([0-9][0-9;]*)*m'),
    re.compile(r'\x1b\[m'),
    re.compile(r'\x1b\[[0-9][A-Z]'),
]

# Match and ANSI colour control sequence and capture the colour bit

RE_DEBUG = re.compile(r'\x1b(\[[0-9][0-9;]*m)')

# Colour conversions for light backgrounds

LIGHT_TABLE = [['1;33', '0;33']]

# Colour conversions for dark backgrounds

DARK_TABLE = []

# Common time formats found in log files

RE_TIME = [
    {'regex': re.compile(r'[1-9][0-9]{3}-[0-9]{2}-[0-9]{2}T[0-9]{2}:[0-9]{2}:[0-9]{2}[.][0-9]+Z?'), 'replace': '{DATE+TIME}'},
    {'regex': re.compile(r'[0-9]{1,2}:[0-9]{1,2}:[0-9]{1,2}(Z|,[0-9]+)'), 'replace': '{TIME}'},
    {'regex': re.compile(r'[0-9]{1,2}:[0-9]{2}:[0-9]{2}(\.[0-9]{1,6})?([+][0-9]{2}:[0-9]{2})?'), 'replace': '{TIME}'},

    {'regex': re.compile(r'[1-9][0-9]{3}/[0-9][0-9]?/[1-9][0-9]'), 'replace': '{DATE}'},
    {'regex': re.compile(r'[1-9][0-9]/[0-9][0-9]?/[1-9][0-9]{3}'), 'replace': '{DATE}'},
    {'regex': re.compile(r'[0-9]{4}-[0-9]{2}-[0-9]{2}'), 'replace': '{DATE}'},
    {'regex': re.compile(r'[0-9]{2}-[0-9]{2}-[0-9]{4}'), 'replace': '{DATE}'},

    {'regex': re.compile(r'[0-9]([.][0-9]*)*\s*(second[s]?)'), 'replace': '{ELAPSED}'},

    {'find': '{DATE} {TIME}', 'replace': '{DATE+TIME}'},
    {'regex': re.compile(r'[0-9]+m *[0-9]+s'), 'replace': '{ELAPSED}'},
]

# SHA values

RE_SHA256 = [
    {'regex': re.compile(r'[0-9a-f]{64}'), 'replace': '{SHA256}'},
]

RE_SHA1 = [
    {'regex': re.compile(r'[0-9a-f]{40}'), 'replace': '{SHA1}'},
]

# AWS ids

RE_AWS = \
    [
        {'regex': re.compile(r'eni-0[0-9a-f]{16}'), 'replace': '{ENI-ID}'},
        {'regex': re.compile(r'ami-0[0-9a-f]{16}'), 'replace': '{AMI-ID}'},
        {'regex': re.compile(r'snap-0[0-9a-f]{16}'), 'replace': '{AMI-SNAP}'},
        {'regex': re.compile(r'vol-0[0-9a-f]{16}'), 'replace': '{AMI-VOL}'},
        {'regex': re.compile(r'sir-[0-9a-z]{8}'), 'replace': '{SPOT-INSTANCE}'},
        {'regex': re.compile(r'i-0[0-9a-f]{16}'), 'replace': '{EC2-ID}'},
        {'regex': re.compile(r'request id: [0-0a-f]{8}-[0-0a-f]{4}-[0-0a-f]{4}-[0-0a-f]{4}-[0-0a-f]{12}'),
         'replace': 'request id: xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx'},
    ]

# Data transfer speeds

RE_SPEED = \
    [
        {'regex': re.compile(r'[0-9.]+ *MB/s'), 'replace': '{SPEED}'},
        {'regex': re.compile(r'[0-9.]+ *MiB/s'), 'replace': '{SPEED}'},
    ]

################################################################################

def regex_replace(data, regexes):
    """ Do a set of regex replacements """

    for regex in regexes:
        if 'find' in regex:
            data = data.replace(regex['find'], regex['replace'])
        elif 'regex' in regex:
            data = regex['regex'].sub(regex['replace'], data)
        else:
            assert False

    return data

################################################################################

def debug_format(data):
    """ Make the ANSI colour codes in the specified string human-readable """

    return RE_DEBUG.sub('{ESC\\1}', data)

################################################################################

def convert_ansi(data, light=True):
    """ Use the conversion table to convert ANSI codes in the string for display
        on a light or dark background """

    table = LIGHT_TABLE if light else DARK_TABLE

    for entry in table:
        data = data.replace('\x1b[%s' % entry[0], '\x1b[%s' % entry[1])

    return data

################################################################################

def remove_times(data):
    """ Attempt to remove obvious time and duration references from a string """

    return regex_replace(data, RE_TIME)

################################################################################

def remove_sha1(data):
    """ Attempt to remove SHA1 references from a string """

    return regex_replace(data, RE_SHA1)

################################################################################

def remove_sha256(data):
    """ Attempt to remove SHA256 references from a string """

    return regex_replace(data, RE_SHA256)

################################################################################

def remove_aws_ids(data):
    """ Attempt to remove a variety of AWS ID references from a string """

    return regex_replace(data, RE_AWS)

################################################################################

def remove_speeds(data):
    """ Attempt to remove data transfer speed references from a string """

    return regex_replace(data, RE_SPEED)

################################################################################

def remove_ansi(text):
    """ Remove ANSI codes from a string """

    if '\x1b' in text:
        for regex in ANSI_REGEXES:
            text = regex.sub('', text)

    return text



################################################################################
""" Pane class for tfm """
################################################################################

import sys
import os
import curses
import fnmatch
import stat
import glob
import time
import threading

from enum import IntEnum

if sys.platform == 'linux':
    import inotify.adapters

import dc_curses
import path
import popup

################################################################################

class SortOrder(IntEnum):
    """ Sort order for filename list """

    FILENAME = 0
    EXTENSION = 1
    MODIFIED_DATE = 2
    SIZE = 3
    NUM_SORTS = 4

SORT_TYPE = ('filename', 'extension', 'modified date', 'size')

################################################################################

def inotify_wait(self):
    """Thread to wait for inotify events and post an event to the queue if there
       any create/delete/modify events in the current directory.
       Sends no more than 1 update per second to avoid drowning the recipient."""

    while True:
        trigger = False
        for event in self.ino.event_gen(yield_nones=False, timeout_s=1):
            (_, events, path, _) = event

            if path == self.current_dir and ('IN_CREATE' in events or 'IN_DELETE' in events or 'IN_MODIFY' in events):
                trigger = True

        if trigger:
            self.event_queue.put(('inotify', self.index))

################################################################################

class Pane():
    """ Class for a file manager pane """

    def __init__(self, index, num_panes, colours, event_queue):
        # Create window for the pane (dummy size and position initially)

        self.screen = curses.newwin(1, 1, 0, 0)

        self.index = index

        self.current_dir = ''

        self.ino = inotify.adapters.Inotify() if sys.platform == 'linux' else None

        self.set_current_dir(os.getcwd())

        self.event_queue = event_queue

        if sys.platform == 'linux':
            inotify_thread = threading.Thread(target=inotify_wait, args=(self,), daemon=True)
            inotify_thread.start()

        # Default sort order

        self.sort_order = SortOrder.FILENAME
        self.reverse_sort = False

        # Set the attributes of the current review (some are initialised
        # when the screen is drawn)

        # Index of the current file in the filtered_file_indices

        self.current = 0

        self.offset = 0
        self.num_panes = num_panes
        self.colours = colours

        self.searchstring = None

        self.height = self.width = -1
        self.file_list_y = 1
        self.file_list_h = -1

        # File list is a list of the files in the current directory

        self.file_list = []

        # Filtered file list is a list of the indices in file_list of the visible files
        # in the current directory

        self.filtered_file_indices = []

        # Set of the names of currently-tagged files

        self.tagged_set = set()

        self.in_filter = self.out_filter = None
        self.hide_hidden_filter = True

        self.file_display_fields = ['size', 'mtime']

        # Set up dircolor highlighting

        self.dircolours = dc_curses.CursesDircolors(reserved=self.colours['reserved_colours'])

        # Generate the list of files to be shown (takes filtering into account)

        self.update_files()

    ################################################################################

    def sort_file_list(self):
        """ Sort the file list according to the current sort order """

        if self.sort_order == SortOrder.FILENAME:
            self.file_list.sort(reverse=self.reverse_sort, key=lambda entry: (not entry['isdir'], os.path.basename(entry['name'])))
        elif self.sort_order == SortOrder.EXTENSION:
            self.file_list.sort(reverse=self.reverse_sort, key=lambda entry: (not entry['isdir'], entry['name'].split('.')[-1]))
        elif self.sort_order == SortOrder.MODIFIED_DATE:
            self.file_list.sort(reverse=self.reverse_sort, key=lambda entry: (not entry['isdir'], entry['mtime']))
        elif self.sort_order == SortOrder.SIZE:
            self.file_list.sort(reverse=self.reverse_sort, key=lambda entry: (not entry['isdir'], entry['size']))

    ################################################################################

    def update_files(self):
        """ Get the list of files
        """

        def file_stats(filename):
            """ Get the stats for a file """

            filestat = os.stat(filename, follow_symlinks=False)

            info = {'name': filename,
                    'mode': filestat.st_mode,
                    'uid': filestat.st_uid,
                    'gid': filestat.st_gid,
                    'size': filestat.st_size,
                    'atime': filestat.st_atime,
                    'mtime': filestat.st_mtime,
                    'ctime': filestat.st_ctime,
                    'isdir': stat.S_ISDIR(filestat.st_mode)}

            return info

        # Rebuild the file list

        self.file_list = []
        for filename in glob.glob(os.path.join(self.current_dir, '*')) + glob.glob(os.path.join(self.current_dir, '.*')):
            self.file_list.append(file_stats(filename))

        # Update the tagged file list to contain only current files

        self.tagged_set = {entry['name'] for entry in self.file_list if entry['name'] in self.tagged_set}

        # Optionally add '..' as an entry

        if self.current_dir != '/':
            self.file_list.append(file_stats('..'))

        self.sort_file_list()
        self.update_file_list()

    ################################################################################

    def update_file_list(self):
        """ Generate the file list from the list of current files with filtering
            applied if enabled """

        self.sort_file_list()

        if self.active_filters():
            self.filtered_file_indices = [i for i, entry in enumerate(self.file_list) if not self.filtered(entry)]
        else:
            self.filtered_file_indices = range(len(self.file_list))

    ################################################################################

    def active_filters(self):
        """ Return true if any filters are active """

        return self.out_filter or \
               self.in_filter or \
               self.hide_hidden_filter

    ################################################################################

    def filtered(self, entry):
        """ Return True if an entry is hidden by one or more filters """

        result = False

        if self.out_filter and fnmatch.fnmatch(entry['name'], self.out_filter):
            result = True

        elif self.in_filter and not fnmatch.fnmatch(entry['name'], self.in_filter):
            result = True

        elif self.hide_hidden_filter:
            base_name = os.path.basename(entry['name'])
            if base_name[0] == '.' and base_name != '..':
                result = True

        return result

    ################################################################################

    def constrain_display_parameters(self):
        """ Ensure that the current display parameters are within range - easier
            to do it in one place for all of them than check individually whenever we
            change any of them """

        self.current = max(min(self.current, len(self.filtered_file_indices) - 1), 0)
        self.offset = min(len(self.filtered_file_indices) - 1, max(0, self.offset))

        # Keep the current entry on-screen

        if self.current >= self.offset + self.height - 2:
            self.offset = self.current
        elif self.current < self.offset:
            self.offset = self.current

    ################################################################################

    def file_info_display(self, filename):
        """ Extract the additional file info fields displayed to the right
            of the filename """

        data = []
        for field in self.file_display_fields:
            if field == 'name':
                data.append(filename['name'])
            elif field in ('atime', 'mtime', 'ctime'):
                data.append(time.strftime('%x %X', time.gmtime(filename[field])))
            elif field == 'uid':
                pass
            elif field == 'gid':
                pass
            elif field == 'mode':
                pass
            elif field == 'size':
                data.append(str(filename[field]))

        return ' '.join(data)

    ################################################################################

    def show_file_list(self, current_pane):
        """ Draw the current page of the file list """

        for ypos in range(0, self.file_list_h):

            normal_colour = curses.color_pair(self.colours['normal'])

            if 0 <= self.offset + ypos < len(self.filtered_file_indices):
                # Work out what colour to render the file details in

                current_file = self.file_list[self.filtered_file_indices[self.offset + ypos]]

                current = self.offset + ypos == self.current

                # The text to render

                filename = os.path.basename(current_file['name'])

                data = self.file_info_display(current_file)

                name = f'/{filename}' if current_file['isdir'] else filename
                name = f'* {name}' if current_file['name'] in self.tagged_set else f'  {name}'

                if len(name) > self.width - len(data):
                    entry = name[:self.width - 3] + '...'
                else:
                    entry = name + ' ' * (self.width - len(name) - len(data)) + data

                file_colour = self.dircolours.get_colour_pair(current_file['name'], current_file['mode'])
            else:
                filename = entry = None
                current = False
                file_colour = normal_colour

            # Reverse the colours if this the cursor line

            if current and current_pane:
                file_colour |= curses.A_REVERSE
                normal_colour |= curses.A_REVERSE

            # Write the prefix, filename, and, if necessary, padding

            self.screen.move(self.file_list_y + ypos, 0)
            if entry:
                self.screen.addstr(entry, file_colour)
            else:
                self.screen.clrtoeol()

            # if len(filename) < self.width:
            #    self.screen.addstr(self.file_list_y + ypos, len(filename), ' ' * (self.width - len(filename)), normal_colour)

        current_dir = path.trimpath(self.current_dir, self.width)

        self.screen.move(0, 0)
        self.screen.attron(curses.color_pair(self.colours['status']))
        self.screen.addstr(current_dir + ' '*(self.width-len(current_dir)))

        self.screen.refresh()

        if not self.filtered_file_indices:
            with popup.PopUp(self.screen, 'All files are hidden - Press \'c\' to clear filters.', self.colours['status']):
                pass

    ################################################################################

    def filter_description(self):
        """ Return a textual description of the active filters """

        filters = []

        if self.out_filter:
            filters.append('filter-out wildcard')

        if self.in_filter:
            filters.append('filter-in wildcard')

        return ', '.join(filters)

    ################################################################################

    def clear_filters(self):
        """ Clear all filters """

        if self.out_filter or self.in_filter:
            self.out_filter = self.in_filter = None
            self.update_file_list()

    ################################################################################

    def reload_changes(self):
        """ Update the list of files in case something external has
            changed it. """

        self.update_files()

    ################################################################################

    def get_current_dir(self):
        """ Get the current directory for the pane """

        return self.current_dir

    ################################################################################

    def set_current_dir(self, directory):
        """ Set the current directory for the pane """

        if self.current_dir and self.ino:
            self.ino.remove_watch(self.current_dir)

        self.current_dir = os.path.normpath(directory)

        if self.ino:
            self.ino.add_watch(directory)

    ################################################################################

    def get_current_file(self):
        """ Get the current file for the pane """

        return self.file_list[self.filtered_file_indices[self.current]]

    ################################################################################

    def get_tagged_files(self):
        """ Get the list of tagged files, or the current file if none are tagged """

        if self.tagged_set:
            return [self.file_list[entry] for entry in self.filtered_file_indices if self.file_list[entry]['name'] in self.tagged_set]

        return [self.get_current_file()]

    ################################################################################

    def search_entry(self, searchstring):
        """ Search for the next match with the specified search string """

        for i in list(range(self.current + 1, len(self.filtered_file_indices))) + list(range(0, self.current)):
            if fnmatch.fnmatch(os.path.basename(self.file_list[self.filtered_file_indices[i]]['name']), searchstring):
                self.current = i
                break

    ################################################################################

    def search_match(self, searchstring):
        """ Search for the first match """

        self.searchstring = searchstring
        self.search_next_match()

    ################################################################################

    def search_next_match(self):
        """ Search for the next match with the current search string """

        self.search_entry(self.searchstring)

    ################################################################################

    def move_end(self):
        """ Move to the end of the file list """

        self.current = len(self.filtered_file_indices) - 1

    ################################################################################

    def move_top(self):
        """ Move to the top of the file list """

        self.current = self.offset = 0

    ################################################################################

    def move_to_file(self, filename):
        """ Move to the specified file (if it exists) in the current directory
            or to the top if not """

        self.current = self.offset = 0
        if filename:
            self.search_entry(filename)

    ################################################################################

    def move(self, delta):
        """ Move up or down the file list """

        self.current += delta

    ################################################################################

    def filter_out(self, filter_out):
        """ Set an exclusion filter """

        self.out_filter = filter_out
        self.in_filter = None
        self.update_file_list()

    ################################################################################

    def filter_in(self, filter_in):
        """ Set an inclusion filter """

        self.in_filter = filter_in
        self.out_filter = None
        self.update_file_list()

    ################################################################################

    def move_page_down(self):
        """ Page down """

        pos = self.current - self.offset
        self.offset += self.file_list_h - 1
        self.current = self.offset + pos

    ################################################################################

    def move_page_up(self):
        """ Page up """

        pos = self.current - self.offset
        self.offset -= self.file_list_h - 1
        self.current = self.offset + pos

    ################################################################################

    def set_sort_order(self, value):
        """ Set the sort order """

        self.sort_order = (self.sort_order + value) % SortOrder.NUM_SORTS

        self.update_sort()

    ################################################################################

    def get_sort_order(self):
        """ Get the current sort order """

        return self.sort_order

    ################################################################################

    def sort_type_msg(self):
        """ Return a textual explanation of the current sort type """

        if self.reverse_sort:
            msg = f'Reverse-sorting by {SORT_TYPE[self.sort_order]}'
        else:
            msg = f'Sorting by {SORT_TYPE[self.sort_order]}'

        return msg

    ################################################################################

    def reverse_sort_order(self):
        """ Reverse the sort order """

        self.reverse_sort = not self.reverse_sort
        self.update_sort()

    ################################################################################

    def update_sort(self):
        """ Update the sort """

        msg = self.sort_type_msg()

        with popup.PopUp(self.screen, msg, self.colours['status']):
            self.update_file_list()

    ################################################################################

    def set_pane_coords(self, y, x, height, width):
        """ Set the pane height given the pane display area """

        pane_width = width//self.num_panes

        self.height = height
        self.file_list_h = height-1
        self.width = pane_width-1  # TODO: Why '-1'?
        self.screen.resize(height, pane_width)
        self.screen.mvwin(y, x + pane_width*self.index)

    ################################################################################

    def tag_current(self):
        """ Tag the current entry (unless it is '..') """

        current = self.file_list[self.filtered_file_indices[self.current]]['name']

        if current != '..':
            if current in self.tagged_set:
                self.tagged_set.remove(current)
            else:
                self.tagged_set.add(current)

    ################################################################################

    def untag(self, wildcard=None):
        """ Tag all, or selected tagged items """

        if wildcard:
            remove_tags = set()
            for entry in self.tagged_set:
                if fnmatch.fnmatch(self.filtered_file_indices[entry], wildcard):
                    remove_tags.add(entry)

                self.tagged_set -= remove_tags
        else:
            self.tagged_set = set()

    ################################################################################

    def get_hidden_visibility(self):
        """ Return the current state of hidden file visibility """

        return not self.hide_hidden_filter

    ################################################################################

    def set_hidden_visibility(self, state=False):
        """ Set the visibility of hidden files """

        self.hide_hidden_filter = not state

        change_txt = 'Hiding' if self.hide_hidden_filter else 'Showing'

        with popup.PopUp(self.screen, f'{change_txt} hidden files', self.colours['status']):
            self.update_file_list()



#! /usr/bin/env python3

################################################################################
""" Colour colour output

    Copyright (C) 2017-18 John Skilleter

    Licence: GPL v3 or later

    0-15 are the standard VGA colour codes
    16-21 are a few extras
    22-231 appear to be a sort of colour cube
    232-255 are 24 shades of grey
"""
################################################################################

import sys
import re

################################################################################
# Constants

_ANSI_NORMAL = '\x1b[0m'

_ANSI_BOLD = '\x1b[1m'
_ANSI_FAINT = '\x1b[2m'
_ANSI_ITALIC = '\x1b[3m'
_ANSI_UNDERSCORE = '\x1b[4m'
_ANSI_BLINK = '\x1b[5m'
_ANSI_REVERSE = '\x1b[7m'
_ANSI_STRIKE = '\x1b[9m'

_ANSI_BLACK = '\x1b[30m'
_ANSI_RED = '\x1b[31m'
_ANSI_GREEN = '\x1b[32m'
_ANSI_YELLOW = '\x1b[33m'
_ANSI_BLUE = '\x1b[34m'
_ANSI_MAGENTA = '\x1b[35m'
_ANSI_CYAN = '\x1b[36m'
_ANSI_WHITE = '\x1b[37m'

_ANSI_BLACK_B = '\x1b[1;30m'
_ANSI_RED_B = '\x1b[1;31m'
_ANSI_GREEN_B = '\x1b[1;32m'
_ANSI_YELLOW_B = '\x1b[1;33m'
_ANSI_BLUE_B = '\x1b[1;34m'
_ANSI_MAGENTA_B = '\x1b[1;35m'
_ANSI_CYAN_B = '\x1b[1;36m'
_ANSI_WHITE_B = '\x1b[1;37m'

_ANSI_BBLACK = '\x1b[40m'
_ANSI_BRED = '\x1b[41m'
_ANSI_BGREEN = '\x1b[42m'
_ANSI_BYELLOW = '\x1b[43m'
_ANSI_BBLUE = '\x1b[44m'
_ANSI_BMAGENTA = '\x1b[45m'
_ANSI_BCYAN = '\x1b[46m'
_ANSI_BWHITE = '\x1b[47m'

_ANSI_BBLACK_B = '\x1b[1;40m'
_ANSI_BRED_B = '\x1b[1;41m'
_ANSI_BGREEN_B = '\x1b[1;42m'
_ANSI_BYELLOW_B = '\x1b[1;43m'
_ANSI_BBLUE_B = '\x1b[1;44m'
_ANSI_BMAGENTA_B = '\x1b[1;45m'
_ANSI_BCYAN_B = '\x1b[1;46m'
_ANSI_BWHITE_B = '\x1b[1;47m'

_ANSI_CLEAREOL = '\x1b[K'

# Looking up tables for converting textual colour codes to ANSI codes

ANSI_REGEXES = \
    (
        (r'\[NORMAL:(.*?)\]', r'%s\1%s' % (_ANSI_NORMAL, _ANSI_NORMAL)),
        (r'\[BOLD:(.*?)\]', r'%s\1%s' % (_ANSI_BOLD, _ANSI_NORMAL)),
        (r'\[FAINT:(.*?)\]', r'%s\1%s' % (_ANSI_FAINT, _ANSI_NORMAL)),
        (r'\[ITALIC:(.*?)\]', r'%s\1%s' % (_ANSI_ITALIC, _ANSI_NORMAL)),
        (r'\[UNDERSCORE:(.*?)\]', r'%s\1%s' % (_ANSI_UNDERSCORE, _ANSI_NORMAL)),
        (r'\[BLINK:(.*?)\]', r'%s\1%s' % (_ANSI_BLINK, _ANSI_NORMAL)),
        (r'\[REVERSE:(.*?)\]', r'%s\1%s' % (_ANSI_REVERSE, _ANSI_NORMAL)),
        (r'\[STRIKE:(.*?)\]', r'%s\1%s' % (_ANSI_STRIKE, _ANSI_NORMAL)),

        (r'\[BLACK:(.*?)\]', r'%s\1%s' % (_ANSI_BLACK, _ANSI_NORMAL)),
        (r'\[RED:(.*?)\]', r'%s\1%s' % (_ANSI_RED, _ANSI_NORMAL)),
        (r'\[GREEN:(.*?)\]', r'%s\1%s' % (_ANSI_GREEN, _ANSI_NORMAL)),
        (r'\[YELLOW:(.*?)\]', r'%s\1%s' % (_ANSI_YELLOW, _ANSI_NORMAL)),
        (r'\[BLUE:(.*?)\]', r'%s\1%s' % (_ANSI_BLUE, _ANSI_NORMAL)),
        (r'\[MAGENTA:(.*?)\]', r'%s\1%s' % (_ANSI_MAGENTA, _ANSI_NORMAL)),
        (r'\[CYAN:(.*?)\]', r'%s\1%s' % (_ANSI_CYAN, _ANSI_NORMAL)),
        (r'\[WHITE:(.*?)\]', r'%s\1%s' % (_ANSI_WHITE, _ANSI_NORMAL)),

        (r'\[BLACK_B:(.*?)\]', r'%s\1%s' % (_ANSI_BLACK_B, _ANSI_NORMAL)),
        (r'\[RED_B:(.*?)\]', r'%s\1%s' % (_ANSI_RED_B, _ANSI_NORMAL)),
        (r'\[GREEN_B:(.*?)\]', r'%s\1%s' % (_ANSI_GREEN_B, _ANSI_NORMAL)),
        (r'\[YELLOW_B:(.*?)\]', r'%s\1%s' % (_ANSI_YELLOW_B, _ANSI_NORMAL)),
        (r'\[BLUE_B:(.*?)\]', r'%s\1%s' % (_ANSI_BLUE_B, _ANSI_NORMAL)),
        (r'\[MAGENTA_B:(.*?)\]', r'%s\1%s' % (_ANSI_MAGENTA_B, _ANSI_NORMAL)),
        (r'\[CYAN_B:(.*?)\]', r'%s\1%s' % (_ANSI_CYAN_B, _ANSI_NORMAL)),
        (r'\[WHITE_B:(.*?)\]', r'%s\1%s' % (_ANSI_WHITE_B, _ANSI_NORMAL)),

        (r'\[BBLACK:(.*?)\]', r'%s\1%s' % (_ANSI_BBLACK, _ANSI_NORMAL)),
        (r'\[BRED:(.*?)\]', r'%s\1%s' % (_ANSI_BRED, _ANSI_NORMAL)),
        (r'\[BGREEN:(.*?)\]', r'%s\1%s' % (_ANSI_BGREEN, _ANSI_NORMAL)),
        (r'\[BYELLOW:(.*?)\]', r'%s\1%s' % (_ANSI_BYELLOW, _ANSI_NORMAL)),
        (r'\[BBLUE:(.*?)\]', r'%s\1%s' % (_ANSI_BBLUE, _ANSI_NORMAL)),
        (r'\[BMAGENTA:(.*?)\]', r'%s\1%s' % (_ANSI_BMAGENTA, _ANSI_NORMAL)),
        (r'\[BCYAN:(.*?)\]', r'%s\1%s' % (_ANSI_BCYAN, _ANSI_NORMAL)),
        (r'\[BWHITE:(.*?)\]', r'%s\1%s' % (_ANSI_BWHITE, _ANSI_NORMAL)),

        (r'\[BBLACK_B:(.*?)\]', r'%s\1%s' % (_ANSI_BBLACK_B, _ANSI_NORMAL)),
        (r'\[BRED_B:(.*?)\]', r'%s\1%s' % (_ANSI_BRED_B, _ANSI_NORMAL)),
        (r'\[BGREEN_B:(.*?)\]', r'%s\1%s' % (_ANSI_BGREEN_B, _ANSI_NORMAL)),
        (r'\[BYELLOW_B:(.*?)\]', r'%s\1%s' % (_ANSI_BYELLOW_B, _ANSI_NORMAL)),
        (r'\[BBLUE_B:(.*?)\]', r'%s\1%s' % (_ANSI_BBLUE_B, _ANSI_NORMAL)),
        (r'\[BMAGENTA_B:(.*?)\]', r'%s\1%s' % (_ANSI_BMAGENTA_B, _ANSI_NORMAL)),
        (r'\[BCYAN_B:(.*?)\]', r'%s\1%s' % (_ANSI_BCYAN_B, _ANSI_NORMAL)),
        (r'\[BWHITE_B:(.*?)\]', r'%s\1%s' % (_ANSI_BWHITE_B, _ANSI_NORMAL)),
    )

ANSI_CODES = \
    (
        ('[NORMAL]', _ANSI_NORMAL),
        ('[BOLD]', _ANSI_BOLD),
        ('[FAINT]', _ANSI_FAINT),
        ('[ITALIC]', _ANSI_ITALIC),
        ('[UNDERSCORE]', _ANSI_UNDERSCORE),
        ('[BLINK]', _ANSI_BLINK),
        ('[REVERSE]', _ANSI_REVERSE),
        ('[STRIKE]', _ANSI_STRIKE),

        ('[BLACK]', _ANSI_BLACK),
        ('[RED]', _ANSI_RED),
        ('[GREEN]', _ANSI_GREEN),
        ('[YELLOW]', _ANSI_YELLOW),
        ('[BLUE]', _ANSI_BLUE),
        ('[MAGENTA]', _ANSI_MAGENTA),
        ('[CYAN]', _ANSI_CYAN),
        ('[WHITE]', _ANSI_WHITE),

        ('[BLACK_B]', _ANSI_BLACK_B),
        ('[RED_B]', _ANSI_RED_B),
        ('[GREEN_B]', _ANSI_GREEN_B),
        ('[YELLOW_B]', _ANSI_YELLOW_B),
        ('[BLUE_B]', _ANSI_BLUE_B),
        ('[MAGENTA_B]', _ANSI_MAGENTA_B),
        ('[CYAN_B]', _ANSI_CYAN_B),
        ('[WHITE_B]', _ANSI_WHITE_B),

        ('[BBLACK]', _ANSI_BBLACK),
        ('[BRED]', _ANSI_BRED),
        ('[BGREEN]', _ANSI_BGREEN),
        ('[BYELLOW]', _ANSI_BYELLOW),
        ('[BBLUE]', _ANSI_BBLUE),
        ('[BMAGENTA]', _ANSI_BMAGENTA),
        ('[BCYAN]', _ANSI_BCYAN),
        ('[BWHITE]', _ANSI_BWHITE),

        ('[BBLACK_B]', _ANSI_BBLACK_B),
        ('[BRED_B]', _ANSI_BRED_B),
        ('[BGREEN_B]', _ANSI_BGREEN_B),
        ('[BYELLOW_B]', _ANSI_BYELLOW_B),
        ('[BBLUE_B]', _ANSI_BBLUE_B),
        ('[BMAGENTA_B]', _ANSI_BMAGENTA_B),
        ('[BCYAN_B]', _ANSI_BCYAN_B),
        ('[BWHITE_B]', _ANSI_BWHITE_B),
    )

# Regex to match an ANSI control sequence

RE_ANSI = re.compile(r'\x1b\[([0-9][0-9;]*)*m')

################################################################################

def format(txt):
    """ Convert textual colour codes in a string to ANSI codes.
        Codes can be specified as either [COLOUR], where all following text
        is output in the specified colour or [COLOR:text] where only 'text' is
        output in the colour, with subsequent text output in the default colours """

    # Replace [COLOUR:text] with COLOURtextNORMAL using regexes

    if re.search(r'\[.*:.*\]', txt):
        for regex in ANSI_REGEXES:
            txt = re.sub(regex[0], regex[1], txt)

    # Replace [COLOUR] with COLOUR

    if re.search(r'\[.*\]', txt):
        for code in ANSI_CODES:
            txt = txt.replace(code[0], code[1])

    # Now replace [N(N)(N)] with 256 colour colour code.

    while True:
        p = re.match(r'.*\[([0-9]{1,3})\].*', txt)
        if p is None:
            break

        value = int(p.group(1))
        txt = txt.replace('[%s]' % p.group(1), '\x1b[38;5;%dm' % value)

    while True:
        p = re.match(r'.*\[B([0-9]{1,3})\].*', txt)
        if p is None:
            break

        value = int(p.group(1))
        txt = txt.replace('[B%s]' % p.group(1), '\x1b[48;5;%dm' % value)

    return txt

################################################################################

def write(txt=None, newline=True, stream=sys.stdout, indent=0, strip=False, cleareol=False, cr=False):
    """ Write to the specified stream (defaulting to stdout), converting colour codes to ANSI
        txt can be None, a string or a list of strings."""

    if txt:
        if isinstance(txt, str):
            txt = txt.split('\n')

        for n, line in enumerate(txt):
            line = format(line)

            if strip:
                line = line.strip()

            if indent:
                stream.write(' ' * indent)

            stream.write(line)

            if cleareol:
                stream.write(_ANSI_CLEAREOL)

            if newline or n < len(txt) - 1:
                stream.write('\n')
            elif cr:
                stream.write('\r')

    else:
        if cleareol:
            stream.write(_ANSI_CLEAREOL)

        if newline:
            stream.write('\n')
        elif cr:
            stream.write('\r')

################################################################################

def error(txt, newline=True, stream=sys.stderr, status=1, prefix=False):
    """ Write an error message to the specified stream (defaulting to
        stderr) and exit with the specified status code (defaulting to 1)
        Prefix the output with 'ERROR:' in red if prefix==True """

    if prefix:
        write('[RED:ERROR]: ', newline=False, stream=stream, )

    write(txt, newline, stream)

    sys.exit(status)

################################################################################

def warning(txt, newline=True, stream=sys.stderr, prefix=False):
    """ Write a warning message to the specified stream (defaulting to
        stderr). Prefix the output with 'WARNING:' in red if prefix==True """

    if prefix:
        write('[RED:WARNING]: ', newline=False, stream=stream, )

    write(txt, newline, stream)

################################################################################

if __name__ == '__main__':
    for combo in (0, 1, 2):
        print()
        if combo == 0:
            print('Background colours')
        elif combo == 1:
            print('Foreground colours')
        else:
            print('Combinations')

        print()
        for y in range(0, 16):
            for x in range(0, 16):
                colour = x + y * 16

                if combo == 0:
                    write(format('[B%d]%4d' % (colour, colour)), newline=False)
                elif combo == 1:
                    write(format('[%d]%4d' % (colour, colour)), newline=False)
                else:
                    write(format('[B%d]%4d[%d]/%4d ' % (colour, colour, 255 - colour, 255 - colour)), newline=False)

            write('[NORMAL]')

    print()

    write('Foreground: [RED]red [GREEN]green [BLACK]black [NORMAL]normal')
    write('Background: [BRED]red [BGREEN]green [BBLACK]black [NORMAL]normal')

    write('Foreground: [BBLUE:blue] [RED:red] normal')

    write('Bright foreground: [RED_B]red [GREEN_B]green [BLACK_B]black [NORMAL]normal')
    write('Bright background: [BRED_B]red [BGREEN_B]green [BBLACK_B]black [NORMAL]normal')

    write('Foreground: [BBLUE:blue_B] [RED:red_B] normal')

    print()

    write('[NORMAL:Normal text]')
    write('[FAINT:Faint text]')
    write('[ITALIC:Italic text]')
    write('[UNDERSCORE:Underscored text]')
    write('[BLINK:Blinking text]')
    write('[REVERSE:Reverse text]')
    write('[STRIKE:Strikethrough text]')

    print()

    error('Error message (nothing should be output after this)', status=0)

    write('This message should not appear')



#! /usr/bin/env python3

################################################################################
""" Thingy file handling functions

    Copyright (C) 2017-18 John Skilleter

    High-level file access functions not provided by the Python libraries
"""
################################################################################

import os
import shutil
import subprocess

################################################################################

def is_binary_file(filename):
    """ Return True if there is a strong likelihood that the specified file
        is binary. """

    return file_type(filename, mime=True).endswith('binary')

################################################################################

def file_type(filename, mime=False):
    """ Return a textual description of the file type """

    # The file utility does not return an error if the file does not exist or
    # is not a file, so we have to do that.

    if not os.path.isfile(filename) or not os.access(filename, os.R_OK):
        raise FileNotFoundError('Unable to access %s' % filename)

    cmd = ['file', '--brief']

    if mime:
        cmd.append('--mime')

    cmd.append(filename)

    result = subprocess.run(cmd, capture_output=True, check=False, text=True)

    return result.stdout.split('\n')[0] if result.returncode==0 else None

################################################################################

def format_size(size, always_suffix=False):
    """ Convert a memory/disk size into appropriately-scaled units in bytes,
        MiB, GiB, TiB as a string """

   # Keep all the maths positive

    if size < 0:
        size = -size
        sign = '-'
    else:
        sign = ''

    # Default divisor and number of decimal places output

    div = 1

    # Step through the multipliers

    for units in (' bytes' if always_suffix else '', ' KiB', ' MiB', ' GiB', ' TiB'):
        # If we can't scale up to this multiplier quit the loop

        if size // div < 1024:
            break

        # Increase the divisor and set the number of decimal places
        # to 1 (set to 0 when we don't have a divisor).

        div *= 1024
    else:
        units = ' PiB'

    # Calculate the size in 10ths so the we get the first digit after
    # the decimal point, doing all the work in the integer domain to
    # avoid rounding errors.

    size_x_10 = (size * 10) // div

    # If the decimal part would be '.0' don't output it

    if size_x_10 % 10 == 0:
        return '%s%d%s' % (sign, size_x_10 // 10, units)

    return '%s%d.%d%s' % (sign, size_x_10 // 10, size_x_10 % 10, units)

################################################################################

def backup(filename, extension='bak', copyfile=True, timestamps=True):
    """ Create a backup of a file by copying or renaming it into a file with extension
        .bak, deleting any existing file with that name.

        Copies the file by default, unless copyfile is False
        Sets the timetamps of the backup file to be the same as the original, unless
        timestamps is False."""

    # Do nothing if the file does not exist

    if not os.path.isfile(filename):
        return

    # Split on the dot characters

    filename_comp = filename.split('.')

    # Replace the extension with the specified on (or 'bak' by
    # default) or add it if the filename did not have an extension.

    if len(filename_comp) > 1:
        filename_comp[-1] = extension
    else:
        filename_comp.append(extension)

    backupname = '.'.join(filename_comp)

    # Remove any existing backup file

    if os.path.isfile(backupname):
        os.unlink(backupname)

    # Create the backup by copying or renaming the file, optionally preserving the
    # timestamp of the original file

    if timestamps:
        file_info = os.stat(filename)

    if copyfile:
        shutil.copyfile(filename, backupname)
    else:
        os.rename(filename, backupname)

    if timestamps:
        os.utime(backupname, ns=(file_info.st_atime_ns, file_info.st_mtime_ns))

################################################################################
# Test code

if __name__ == "__main__":
    print('Is /bin/sh binary:      %s' % is_binary_file('/bin/sh'))
    print('Is files.py binary:     %s' % is_binary_file('files.py'))
    print('')

    for mimeflag in (False, True):
        print('/bin/sh is:             %s' % file_type('/bin/sh', mimeflag))
        print('/bin/dash is:           %s' % file_type('/bin/dash', mimeflag))
        print('')

    for sizevalue in (0, 1, 999, 1024, 1025, 1.3 * 1024, 2**32 - 1, 2**64 + 2**49):
        print('%24d is %s' % (sizevalue, format_size(sizevalue)))



#! /usr/bin/env python3

################################################################################
""" GitLab module - implemented using the REST API as some features are not
    available (or don't work) in the official Python module

    Copyright (C) 2017-20 John Skilleter

    Licence: GPL v3 or later

    Note: There are two types of function for returning data from GitLab;
    the paged functions and the non-paged ones - the paged ones return a page
    (normally 20 items) of data and need to be called repeated until no data is
    left whereas the non-paged ones query all the data and concatenate it
    together.

    The paged functions expect a full request string with the URL, as returned
    by the request_string() member. The non-paged ones call request_string()
    to add the URL & API prefix.
"""
################################################################################

import sys
import os

try:
    import requests
except ModuleNotFoundError:
    sys.stderr.write('This code requires the Python "requests" module which should be installed via your package manager\n')
    sys.exit(1)

################################################################################

class GitLabError(Exception):
    """ Gitlab exceptions """

    def __init__(self, response):
        """ Save the error code and text """

        self.status = response.status_code
        self.message = response.reason

    def __str__(self):
        """ Return a string version of the exception """

        return '%s: %s' % (self.status, self.message)

################################################################################

class GitLab:
    """ Class for GitLab access """

    def __init__(self, gitlab, token=None):
        """ Initialisation """

        # Save the GitLab URL

        self.gitlab = gitlab

        # If we have a private token use it, otherwise try and get it from
        # the environmnet

        self.token = token if token else os.getenv('GITLAB_TOKEN', None)

        # Create the default header for requests

        self.header = {'Private-Token': self.token}

    ################################################################################

    @staticmethod
    def encode_project(name):
        """ Encode a project name in the form request by GitLab requests """

        return name.replace('/', '%2F')

    ################################################################################

    def request_string(self, request):
        """ Add the URL/API header onto a request string """

        return '%s/api/v4/%s' % (self.gitlab, request)

    ################################################################################

    def request(self, request, parameters=None):
        """ Send a request to GitLab - handles pagination and returns all the
            results concatenated together """

        if parameters:
            request = '%s?%s' % (request, '&'.join(parameters))

        gl_request = self.request_string(request)

        # Keep requesting data until there's no 'next' link in the response

        while True:
            response = requests.get(gl_request, headers=self.header)

            if not response:
                raise GitLabError(response)

            yield response.json()

            if 'next' not in response.links:
                break

            gl_request = response.links['next']['url']

    ################################################################################

    def paged_request(self, request):
        """ Send a request to GitLab - returns all the results concatenated together
            and returns a page of results along with the request for the next page of
            results (if any).

            Note that the request parameter is the full request string as returned by
            request_string(). """

        response = requests.get(request, headers=self.header)

        result = response.json()

        if not response:
            raise GitLabError(response)

        request = response.links['next']['url'] if 'next' in response.links else None

        return result, request

    ################################################################################

    def projects(self):
        """ Return a list of projects """

        return self.request('projects')

    ################################################################################

    def branches(self, repo):
        """ Return the list of branches in a repo """

        for batch in self.request('projects/%s/repository/branches' % self.encode_project(repo)):
            for branch in batch:
                yield branch

    ################################################################################

    def merge_requests(self, **kwargs):
        """ Return a list of merge requests filtered according to the parameters """

        request = 'merge_requests'

        parameters = []

        for data in kwargs:
            parameters.append('%s=%s' % (data, kwargs[data]))

        for result in self.request(request, parameters):
            for r in result:
                yield r

    ################################################################################

    def default_branch(self, repo):
        """ Query gitlab to retreive the default branch for the repo """

        # Look for the default branch

        for branch in self.branches(repo):
            if branch['default']:
                return branch['name']

        return None

    ################################################################################

    def isbranch(self, repo, branchname):
        """ Return True if the branch exists in the repo """

        request = self.request_string('projects/%s/repository/branches' % self.encode_project(repo))

        while True:
            branches, request = self.paged_request(request)

            for branch in branches:
                if branch['name'] == branchname:
                    return True

            if not request:
                break

        return False



#! /usr/bin/env python3

################################################################################
""" Docker interface for Thingy

    Copyright (C) 2017 John Skilleter

    Note that this:
        * Only implements functions required by docker-purge
        * Only has basic error checking, in that it raises DockerError
          for any error returned by the external docker command.
"""
################################################################################

import subprocess

################################################################################

class DockerError(Exception):
    """ Exception for dockery things """

    pass

################################################################################

def instances(all=False):
    """ Return a list of all current Docker instances """

    cmd = ['docker', 'ps', '-q']

    if all:
        cmd.append('-a')

    instances_list = []
    try:
        process = subprocess.run(cmd, capture_output=True, check=True, text=True)

        for result in process.stdout:
            instances_list.append(result)

    except subprocess.CalledProcessError as exc:
        raise DockerError(exc)

    return instances_list

################################################################################

def stop(instance, force=False):
    """ Stop the specified Docker instance """

    # TODO: force option not implemented

    try:
        subprocess.run(['docker', 'stop', instance], check=True, capture_output=False)

    except suprocess.CalledProcessError as exc:
        raise DockerError(exc)

################################################################################

def rm(instance, force=False):
    """ Remove the specified instance """

    cmd = ['docker', 'rm']

    if force:
        cmd.append('--force')

    cmd.append(instance)

    try:
        subprocess.run(cmd, check=True, capture_output=False)

    except subprocess.CalledProcessError as exc:
        raise DockerError(exc)

################################################################################

def images():
    """ Return a list of all current Docker images """

    try:
        process = subprocess.run(['docker', 'images', '-q'], capture_output=True, check=True)

        for result in process:
            yield result

    except subprocess.CalledProcessError as exc:
        raise DockerError(exc)

################################################################################

def rmi(image, force=False):
    """ Remove the specified image """

    cmd = ['docker', 'rmi']
    if force:
        cmd.append('--force')

    cmd.append(image)

    try:
        subprocess.run(cmd, capture_output=False, check=True)
    except subprocess.CalledProcessError as exc:
        raise DockerError(exc)
